""""""
import os
from pathlib import Path
from .tools import load_json, eprint
from askGPT import DATA_PATH
import toml
from .api.openai import ChatGPT

basicConfig = dict()
basicConfig["userPrompt"] = basicConfig.get("userPrompt"," Human: ")
basicConfig["aiPrompt"] = basicConfig.get("aiPrompt"," AI: ")
basicConfig["maxTokens"] = basicConfig.get("maxTokens","150")
basicConfig["model"] = basicConfig.get("model","text-davinci-003")
basicConfig["temperature"] = basicConfig.get("temperature","0.0")
basicConfig["topP"] = basicConfig.get("topP","1")
basicConfig["frequencyPenalty"] = basicConfig.get("frequencyPenalty","0.5")
basicConfig["presencePenalty"] = basicConfig.get("presencePenalty","0.5")
basicConfig["showDisclaimer"] = basicConfig.get("showDisclaimer",True)
basicConfig["maxRetries"] = basicConfig.get("maxRetries",3)
basicConfig["retryDelay"] = basicConfig.get("retryDelay",15.0)
basicConfig["retryMaxDelay"] = basicConfig.get("retryMaxDelay",60)
basicConfig["retryMultiplier"] = basicConfig.get("retryMultiplier",2)

class Config(object):
    def __init__(self):
        self.rate_limit_per_minute = 20
        self.delay = 60.0 / self.rate_limit_per_minute
        self.disclaimer = "Disclaimer: The advice provided by askGPT is intended for informational and entertainment purposes only. It should not be used as a substitute for professional advice, and we cannot be held liable for any damages or losses arising from the use of the advice provided by askGPT."
        self.settingsPath=os.path.join(os.getenv("HOME"), ".askGPT")
        self.progConfig = dict()
        self.sessionConfig = dict()
        self.credentials = None
        self.has = dict()
        self.has["license"] = False
        self.conversations_path=os.path.join(self.settingsPath, "conversations")
        Path(self.conversations_path).mkdir(parents=True, exist_ok=True)
        self.loadScenarios()
        self.fileExtention=".ai.txt"
        self.loadDefaults()
        self.loadProgConfig()
        self.chat = ChatGPT(self)
        self.chat.loadLicense()
        self.version="0.4.12"
        self.exitCode = 0

    def loadProgConfig(self):
        if os.path.isfile(os.path.join(self.settingsPath, "config.toml")):
            tomlConfig = toml.load(os.path.join(self.settingsPath,"config.toml"))
            self.progConfig.update(tomlConfig["default"])
            print(self.progConfig)
        else:
            self.saveConfig()

    def updateParameter(self,key, val):
        if val == "true":
            val = True
        elif val == "false":
            val = False
        elif val.isnumeric():
            val = int(val)
        elif val.replace(".","",1).isnumeric():
            val = float(val)
        if key in self.sessionConfig: # order matters
            if self.sessionConfig[key] != val:
                # prin  t(f"{key}] = {val}")
                self.sessionConfig[key] = val
        elif key in self.progConfig:
            if self.progConfig[key] != val:
                # print(f"{key} = {val}")
                self.progConfig[key] = val
        


    def saveConfig(self):
        """Save the configuration file"""
        jsonConfig = {'name':'askGPT','default':self.progConfig}
        with open(os.path.join(self.settingsPath,"config.toml"), 'w') as f:
            toml.dump(jsonConfig,f)
        self.update()

    def reloadConfig(self):
        """Reload the configuration file"""
        self.update()

    def get_list(self):
        """
        list the previous conversations saved by askGPT."""
        conv_array = list()
        for line in os.listdir(self.conversations_path):
            if (not line.startswith("."))  and line.endswith(self.fileExtention) and (os.path.isfile(os.path.join(self.conversations_path,line))):
                conv_array.append(line.replace(self.fileExtention,""))
        return sorted(conv_array)

    def loadScenarios(self):
        """if there is not a file named scenarios.json, create it ad add the Neutral scenario"""
        if not os.path.isfile(os.path.join(self.settingsPath,"scenarios.json")):
            # copy the file from PATH
            with open(os.path.join(DATA_PATH,"scenarios.json"), "r") as f:
                data = f.read()
            with open(os.path.join(self.settingsPath,"scenarios.json"), "w") as f:
                f.write(data)
        self.scenarios = load_json(os.path.join(self.settingsPath,"scenarios.json"))


    def loadDefaults(self):
        self.progConfig["userPrompt"] = self.progConfig.get("userPrompt"," Human: ")
        self.progConfig["aiPrompt"] = self.progConfig.get("aiPrompt"," AI: ")
        self.progConfig["maxTokens"] = self.progConfig.get("maxTokens",150)
        self.progConfig["model"] = self.progConfig.get("model","text-davinci-003")
        self.progConfig["temperature"] = self.progConfig.get("temperature",0.0)
        self.progConfig["topP"] = self.progConfig.get("topP",1)
        self.progConfig["frequencyPenalty"] = self.progConfig.get("frequencyPenalty",0.0)
        self.progConfig["presencePenalty"] = self.progConfig.get("presencePenalty",0.0)
        self.progConfig["showDisclaimer"] = self.progConfig.get("showDisclaimer",True)
        self.progConfig["maxRetries"] = self.progConfig.get("maxRetries",3)
        self.progConfig["retryDelay"] = self.progConfig.get("retryDelay",15.0)
        self.progConfig["retryMaxDelay"] = self.progConfig.get("retryMaxDelay",60)
        self.progConfig["retryMultiplier"] = self.progConfig.get("retryMultiplier",2)

    def printConfig(self):
        """Print the configuration file"""
        print(toml.dumps(self.progConfig))

    def update(self):
        """
Load the configuration file from ~/.askGPT/config.toml"""
        self.loadProgConfig()



#!/usr/bin/env python
from click.testing import CliRunner
from askGPT import cli

def test_cli():
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "Show this message and exit." in result.output
    
def test_version():
    runner = CliRunner()
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert "0.4.12" in result.output

def test_show():
    runner = CliRunner()
    result = runner.invoke(cli, ["show", "config"])
    assert result.exit_code == 0
    assert "maxTokens" in result.output
    result = runner.invoke(cli, ["show", "scenarios"])
    assert result.exit_code == 0
    assert "scenarios" in result.output
    result = runner.invoke(cli, ["show", "subjects"])
    assert result.exit_code == 0
    assert "subjects" in result.output
    result = runner.invoke(cli, ["show", "conversations"])
    assert result.exit_code == 0 #FIXME it should be 1
    assert "Please specify what to show" in result.output
    result = runner.invoke(cli, ["show", "subject", "test"])
    assert result.exit_code == 0
    assert "Human: are you still working?" in result.output

def test_config():
    runner = CliRunner()
    result = runner.invoke(cli, ["config", "set", "maxTokens", "10"])
    assert result.exit_code == 2
    result = runner.invoke(cli, [ "config", "--max-tokens", "10"])
    assert result.exit_code == 0
    result = runner.invoke(cli, ["config"])
    assert result.exit_code == 0
    assert "maxTokens = 10" in result.output 


if __name__ == "__main__":
    test_cli()
    test_version()
    test_show()
    test_config()


import cmd
import shlex 
from .tools   import eprint, sanitizeName
import toml
import os
import sys
import rich
from pathlib import Path
from rich import print
from rich.text import Text
from rich.style import Style
from rich.prompt import Prompt, Confirm

danger_style = Style(color="red", blink=False, bold=True)
attention_style = Style(color="yellow", blink=False, bold=True)
ok_style = Style(color="green", blink=False, bold=False)
from rich.console import Console
console = Console()

"""Here we will define the class Shell which is a child of cmd.cmd which will allow us to run all the commands interactively such as query, config, edit."""
class Shell(cmd.Cmd):
    def __init__(self, config) -> None:
        super().__init__()
        
        self.intro = "Welcome to askGPT. Type help or ? to list commands."
        self.doc_header = "Commands (type help <topic>):"
        self.misc_header = "Miscellaneous help topics:"
        self.undoc_header = "Undocumented commands:"
        self.ruler = "-"
        self._config = config
        self.commands = dict()
        self._register_commands()
        self.conversation_parameters = {
            "subject": "test",
            "scenario": "Neutral",
            "model": "text-davinci-003",
            "defaultCommand": "query"

        }
        if os.path.exists(os.path.join(self._config.settingsPath, "last.toml")):
            self.conversation_parameters = toml.load(os.path.join(self._config.settingsPath, "last.toml"))
        self.prompt = f"{self.conversation_parameters['scenario']}> "

    def do_clone(self,args):
        """if len(arg) == 1 then copy the current conversation to a new file using arg[1]"""
        args = shlex.split(args)
        if len(args) == 0:
            eprint("You must provide a name for the new conversation")
        else:
            if len(args) == 1:
                new_subject = sanitizeName(args[0])
            else:
                eprint("You can only provide one argument")
                return
            if new_subject in self._config.get_list():
                eprint(f"Subject {new_subject} already exists")
                return
            current_subject = self.conversation_parameters["subject"]
            filename = os.path.join(self._config.conversations_path, f"{current_subject}{self._config.fileExtention}")
            with open(filename, "r") as r:
                text = r.read()
            filename = os.path.join(self._config.conversations_path, f"{new_subject}{self._config.fileExtention}")
            with open(filename, "w") as w:
                w.write(text)
            self.conversation_parameters["subject"] = new_subject
            self.prompt = f"{new_subject}> "
            print(f"Conversation {new_subject} created")


    def do_recap(self, args):
        """prints the text from the conversation file using the subject.ai.txt """
        args = shlex.split(args)
        if len(args) == 0:
            subject = self.conversation_parameters["subject"]
        else:
            subject = args[0]
            if subject not in self._config.get_list():
                eprint(f"Subject {subject} not found")
                return
        filename = os.path.join(self._config.conversations_path, f"{subject}{self._config.fileExtention}")
        if os.path.isfile(filename):
            with open(filename, "r") as f:
                print(f.read())
        else:
            eprint(f"File {filename} not found")

    def do_greetings(self, args):
        """if args is one of the scenarios, print the greeting of that scenario"""
        args = shlex.split(args)
        if len(args) == 0:
            """print current scenario from conversation_parameters"""
            scenario = self.conversation_parameters["scenario"]
            print(self._config.scenarios[scenario]["greetings"])
            return
        scenario = args[0]
        if scenario in self._config.scenarios:
            print(self._config.scenarios[scenario]["greetings"])
        else:
            eprint(f"Scenario {scenario} not found")

    def do_exec(self,args):
        """Execute a the rest of the line in a bash shell and print the output"""
        args = shlex.split(args)
        if len(args) == 0:
            eprint("No command provided")
            return
        command = " ".join(args)
        os.system(command)


    def do_set(self,args):
        """Set conversation_parameters checking that the keys exist and that the values are valid"""
        args = shlex.split(args)
        if len(args) == 0:
            for key in self.conversation_parameters:
                print(f"{key} = {self.conversation_parameters[key]}")
            return
        if len(args) == 2:
            key = args[0]
            val = args[1]
            if key in self.conversation_parameters:
                if key == "scenario":
                    if val in self._config.scenarios:
                        self.conversation_parameters[key] = val
                        self.prompt = f"{val}> "
                    else:
                        eprint("Scenario not found")
                elif key == "model":
                    if val  in self._config.chat.listModels():
                        self.conversation_parameters[key] = val
                    else:
                        eprint("Model not found")
                else:
                    self.conversation_parameters[key] = val
            elif key in self._config.progConfig:
                if key == "fileExtention":
                    if val[0] != ".":
                        val = "." + val
                self._config.progConfig[key] = val
                if val.isnumeric():
                    val = int(val)
                    self._config.progConfig[key] = val
                elif val.replace(".","").isnumeric():
                    val = float(val)
                    self._config.progConfig[key] = val
                else:
                    self._config.progConfig[key] = val
                self._config.saveConfig()
            else:
                eprint("Key not found")
        else:
            eprint("Wrong number of arguments")
    
    def do_version(self,args):
        """Print the version of the program"""
        print(f"Version {self._config.version}")

    def _register_commands(self):
        """Register all the commands."""
        for name in dir(self):
            if name.startswith("do_"):
                self.commands[name[3:]] = getattr(self, name)

        
    def do_credentials(self, args):
        """credentials: show the credentials."""
        if self._config.credentials:
            print(self._config.credentials)
            if not Confirm.ask("Would you like to replace them?"):
                return
        """ask if to replace"""
        api_key = ""
        while(api_key == ""):
            api_key = Prompt.ask("Enter your openai API key",default="" if self._config.credentials is None else self._config.credentials.split(":")[0])
        
        api_organization=""
        while(api_organization == ""):
            api_organization = Prompt.ask("Enter your openai otganization",default="" if self._config.credentials is None else self._config.credentials.split(":")[1])
        self._config.credentials = f"{api_key}:{api_organization}"

        """if yes, ask for the new credentials"""
        self._config.chat.saveLicense(api_key, api_organization)
        self._config.chat.loadLicense()
        """if no, do nothing"""

    def do_delete(self, subject):
        """delete: delete a subject."""
        subject = sanitizeName(subject)
        if subject in self._config.subjects:
            self._config.subjects.remove(subject)
            os.remove(os.path.join(self._config.conversations_path, subject + self._config.fileExtention))
            
        else:
            eprint("Subject not found")

    def do_edit(self, args):
        """edit: edit a subject."""
        if len(args) == 0:
            subject = self.conversation_parameters["subject"]
        else:
            subject = sanitizeName(args[0])
        if subject in self._config.get_list():
            self._config.chat.editDialog(subject)
        else:
            eprint("Subject not found")

    def do_config(self, args):
        """needed two parameters, the first one is the key and the second one is the value"""
        args = shlex.split(args)
        clean = True
        if len(args) == 0:
            self._config.printConfig()
            return
        elif len(args) == 1:
            if args[1] == "save":
                self._config.saveConfig()
                return
            eprint("No value provided.")
            return
        elif len(args) == 2:
            self._config.updateParameter(args[0], args[1])
        else:
            eprint("Unrecognized parameter.")
        if not clean:
            pass
            
    

    def do_submit(self, args):
        """submit: submit a subject."""
        if self._config.has.get("license", False):
            self._config.chat.submitDialog(self.conversation_parameters["subject"], self.conversation_parameters["scenario"])
        else: 
            self._config.chat.loadLicense()
        return

        
    def do_query(self, enquiry, max_tokens: int = 150, temperature: float = 0.9, top_p: float = 1, frequency_penalty: float = 0, presence_penalty: float = 0, stop: list = ["\n", " Human:", " AI:"]):
        """Query the model with the given prompt."""
        """query: query the model with the given prompt.
         <prompt> """
        if not self._config.has.get("license", False):
            self._config.chat.loadLicense()
            return
        if self._config.has.get("license", False):
            with console.status("waiting for response ...", spinner="dots"):
                response = self._config.chat.query(self.conversation_parameters["subject"], self.conversation_parameters["scenario"], enquiry)
                text = Text(response["choices"][0]["text"])
                text.stylize("bold magenta")
            console.print(text)
            """Query the model with the given prompt."""

    def complete_query(self,text, line, begidx, endidx):
        """complete_query: complete the query command."""
        if not text:
            completions = self._config.subjects[:]
        else:
            completions = [ f
                            for f in self._config.subjects
                            if f.startswith(text)
                            ]
        return completions

    def do_show(self, arg):
        """
        show: show the config|scenarios|models|subjects or the conversation inside a subject.
        <config|scenarios|models|subjects|subject <subject>>"""
        args = shlex.split(arg)
        if len(args) == 0:
            eprint("Show config|scenarios|models|subjects or the conversation inside a subject.")
            return
        elif len(args) == 1:
            if args[0] == "config":
                print("Current configuration:")
                print(toml.dumps(self._config.progConfig))
                return
            elif args[0] == "scenarios":
                print("Current scenarios:")
                for scenario in self._config.scenarios.keys():
                    print(scenario)
                return
            elif args[0] == "subjects":
                print("Current subjects:")
                for subject in self._config.get_list():
                    print(subject)
                return
            elif args[0] == "models":
                if self._config.has.get("license", False):
                    print("Current models:")
                    for val  in self._config.chat.listModels():
                        print(val)
                else: 
                    self._config.chat.loadLicense()
                return
            else:
                eprint("Unrecognized parameter.")
                return
        elif len(args) == 2:
            if args[0] == "subject":
                subject = sanitizeName(args[1])
                eprint("Show the conversation inside a subject.")
                if os.path.isfile(os.path.join(self._config.settingsPath, 'conversations', subject + self._config.fileExtention)):
                    with open(os.path.join(self._config.settingsPath, 'conversations', subject + self._config.fileExtention), 'r') as f:
                        print(f.read())
                        return
            else:
                eprint("Unrecognized parameter.")
                return
        else:
            eprint("Unrecognized parameter.")
            return
            
        """config: configure the model."""
        print("configuring the model...")
        print("TODO: implement the config function.")

    def saveSession(self):
        """EOF: exit the shell."""
        """ save the current parameters in a toml file to be loaded in the next session"""
        with open(os.path.join(os.path.join(self._config.settingsPath, "last.toml")), "w") as f:
            toml.dump(self.conversation_parameters, f)


    def do_EOF(self, arg):
        self.saveSession()
        return True

    def do_exit(self, arg):
        """exit: exit the shell."""
        self.saveSession()
        return True

    def do_quit(self, arg):
        """quit: exit the shell."""
        self.saveSession()
        return True
    
    def do_help(self, arg):
        """help: show the help message."""
        if arg:
            try:
                print(getattr(self, 'do_' + arg).__doc__)
            except AttributeError:
                print(self.nohelp % (arg,))
        else:
            names = self.commands.keys()
            cmds_doc = []
            cmds_undoc = []
            cmds_misc = []
            for name in names:
                if name[:5] == 'help_':
                    cmds_misc.append(name[5:])
                elif getattr(self, 'do_' + name).__doc__:
                    cmds_doc.append(name)
                else:
                    cmds_undoc.append(name)
            self.print_topics(self.doc_header, cmds_doc, 15, 80)
            self.print_topics(self.misc_header, cmds_misc, 15, 80)
            self.print_topics(self.undoc_header, cmds_undoc, 15, 80)

    def emptyline(self):
        """emptyline: do nothing."""
        pass

    def default(self, line):
        if line.startswith("!"):
            self.do_exec(line[1:])
        elif self.conversation_parameters.get("defaultCommand","query") == "query":

            self.do_query(line)
        else:
            """default: print the error message."""
            print("Unrecognized command:", line)

    def postloop(self):
        """postloop: print the exit message."""
        print("Exiting askGPT.")

    def precmd(self, line):
        """precmd: print the command."""
        #print("Command:", line)
        return line

    def postcmd(self, stop, line):
        """postcmd: print the result."""
        #print("Result:", stop)
        return stop

    



#!/usr/bin/env python
"""
askGPT is a simple command line tool for interacting with OpenAI's API.
It is based on the example provided by OpenAI at https://beta.openai.com/docs/developer-quickstart/1-introduction
Date: 2023/01/23
"""
__title__ = 'askGPT'
__author__ = 'Meir Michanie'
__license__ = 'MIT'
__credits__ = ''
__version__ = "0.4.12"


import os
from .api.openai import ChatGPT
from .config import Config, basicConfig
import click
from rich import print
import backoff
import time
import toml
import platform
import subprocess
from .shell import Shell
from .tools import eprint, sanitizeName
# from rich.console import Console
# console = Console()

# use pyreadline3 instead of readline on windows
is_windows = platform.system() == "Windows"
if is_windows:
    import pyreadline3  # noqa: F401
else:
    import readline

pass_config = click.make_pass_decorator(Config, ensure=True)

@click.group()
@click.version_option(__version__)
@pass_config
def cli(config):
    if config.progConfig.get("showDisclaimer",True):
        print(config.disclaimer)

@cli.command()
@pass_config
def disclaimer(config):
    """Show the disclaimer"""
    if not config.progConfig.get("showDisclaimer",False):
        print(config.disclaimer)

@cli.command()
@pass_config

def shell(config):
    """Use the cmd module to create an interactive shell where the user can all the commands such as query, edit, config, show. We will call a class which we will write later as a child of cmd.cmd"""
    shell = Shell(config)
    shell.cmdloop()



@cli.command()
@pass_config
@click.option("--set", help="configuration to change")
@click.option("--value", help="AI prompt")
def config(config, set, value):
    """
Change config values"""
    if set:
        if set in config.progConfig:
            config.updateParameter(set, value)
        else:
            print(f"Can't set {set} to {value}. {set} is not a valid configuration value")
    jsonConfig = {'name':'askGPT','default':config.progConfig}
    print(toml.dumps(jsonConfig))
    with open(os.path.join(config.settingsPath,"config.toml"), 'w') as f:
        toml.dump(jsonConfig,f)
        

@cli.command()
@click.option("--subject", "-s", default="quest1", help="The subject of the query. This is used to determine the model to use.")
@click.option("--scenario", "-sc", default="Zork", help="The scenario of the query. This is used to determine the model to use.")
@click.option("--prompt", "-p", default="You are in a maze of twisty little passages, all alike.", help="The prompt to use for the query.")
@click.option("--max_tokens", "-m", default=64, help="The maximum number of tokens to return.")
@pass_config
def query(config, subject, scenario, prompt, max_tokens):
    """Query the model for a response."""
    print(config.chatGPT.query(subject, scenario, prompt, max_tokens))


@cli.command()
@pass_config
@click.option("--subject", prompt="Subject", help="Subject to use to save the conversation")
@click.option("--submit", is_flag=True, help="Submit the dialog")
@click.option("--scenario", default="Neutral", help="scenario to use in the conversation")

def edit(config,subject,submit, scenario):
    """
Edit a conversation"""
    subject = sanitizeName(subject)
    lines = list()
    if os.path.isfile(os.path.join(config.conversations_path, subject + config.fileExtention)):
        with open(os.path.join(config.conversations_path, subject + config.fileExtention), "r") as f:
            lines = f.readlines()
    lines = click.edit("".join(lines))
    if lines is not None:
        with open(os.path.join(config.conversations_path, subject + config.fileExtention), "w") as f:
            f.write(lines)
    if submit:
        config.chat.submitDialog(config,subject, scenario)

@cli.command()
@pass_config
@click.option("--subject", prompt="Subject", help="Subject to use to save the conversation")
@click.option("--scenario", default="Neutral", help="scenario to use in the conversation")
@click.option("--temperature", default=basicConfig["temperature"], type=float, help="Set alternative temperature")
def train(config, subject, scenario, temperature):
    """Train the model with the conversation"""
    config.progConfig["temperature"] = temperature
    config.chat.submitDialog(config, subject, scenario)

@cli.command()
@pass_config
@click.option("--subject", prompt="Subject", help="Subject to use to save the conversation")
@click.option("--scenario", default="Neutral", help="scenario to use in the conversation")
@click.option("--temperature", default=basicConfig["temperature"], type=float, help="Set alternative temperature")
def submit(config, subject, scenario, temperature):
    """Submit without editing the scenario file to openAi api and print out the response."""
    config.progConfig["temperature"] = temperature
    config.chat.submitDialog(config, subject, scenario)




@cli.command()
@pass_config
@click.argument("whatToShow", default="config")
@click.argument("subject", default="" )
def show(config, whattoshow, subject):
    """Show config|scenarios|subjects or the conversation inside a subject."""
    if subject == "":
        if whattoshow == "config":
            print("Current configuration:")
            print(toml.dumps(config.progConfig))
        elif whattoshow == "subjects":
            print("Current subjects:")
            for subject in config.get_list():
                    print(subject)
        elif whattoshow == 'scenarios':
            print("Current scenarios:")
            for scenario in config.scenarios.keys():
                print(scenario)
        elif whattoshow == 'models':
            print("Current models:")
            for model in sorted(list(map(lambda n: n.id,config.chat.Model.list().data))):
                print(model)
        else:
            print("Please specify what to show. Valid options are: config, subjects, scenarios, subject")
            print("In case of passing the option 'subject' please pass as well the subject's name")
    else:
        if os.path.isfile(os.path.join(config.settingsPath, 'conversations', subject + config.fileExtention)):
            with open(os.path.join(config.settingsPath, 'conversations', subject + config.fileExtention), 'r') as f:
                print(f.read())
        else:
            eprint("Subject not found")


@cli.command()
@pass_config
def credentials(config):
    """
Save the API keys to query OpenAI"""
    print("Welcome to askGPT")
    print("Please provide your OpenAI API key and organization")
    print("You can find these values at https://beta.openai.com/account/api-keys")
    config.chat.api_key = input("API_KEY:")
    config.chat.organization = input("ORGANIZATION:")
    with open(os.path.join(config.settingsPath, "credentials"), "w") as f:
        f.write(config.chat.api_key + ":" + config.chat.organization)
    print("askGPT is now ready to use")


@pass_config
@cli.command()
@pass_config
@click.option("--subject", help="Subject of the conversation")
@click.option("--all/--single",  default=False, help="Delete all archived conversations")
def delete(config, subject, all):
    """
Delete the previous conversations saved by askGPT"""
    if all:
        for subject in config.get_list():
            os.remove(os.path.join(config.conversations_path, subject + config.fileExtention))
    elif subject and os.path.isfile(os.path.join(config.conversations_path,sanitizeName(subject) + config.fileExtention)):
        os.remove(os.path.join(config.conversations_path, subject + config.fileExtention))
    else:
        eprint("No chat history with that subject")
        return

@cli.command()
@pass_config
@click.option("--subject", prompt="Subject", help="Subject of the conversation")
@click.option("--enquiry", prompt="Enquiry", help="Your question")
@click.option("--scenario", default="Neutral", help="scenario to use in the conversation")
@click.option("--model", default=basicConfig["model"], help="Set alternative model")
@click.option("--temperature", default=basicConfig["temperature"], type=float, help="Set alternative temperature")
@click.option("--top-p", default=basicConfig["topP"], type=int, help="Set alternative topP")
@click.option("--frequency-penalty", default=basicConfig["frequencyPenalty"], type=float, help="Set alternative frequencyPenalty")
@click.option("--presence-penalty", default=basicConfig["presencePenalty"], type=float, help="Set alternative presencePenalty")
@click.option("--max-tokens", default=basicConfig["maxTokens"], type=int, help="Set alternative maxTokens")
@click.option("--verbose", is_flag=True, help="Show verbose output or just the answer")
@click.option("--save/--no-save", default=True, help="Save the conversation")
@click.option("--retry", is_flag=True, help="In case of error retry the post.")
@click.option("--execute", is_flag=True, help="Parse the AI response and execute it")
def query(config, subject, enquiry, scenario,model, temperature,max_tokens, top_p,  frequency_penalty, presence_penalty, verbose, save, retry, execute): 
    """
Query the OpenAI API with the provided subject and enquiry"""
    enquiry = config.progConfig["userPrompt"] + enquiry
    if subject:
        with open(os.path.join(config.conversations_path, sanitizeName(subject) + config.fileExtention), "a") as f:
            pass
        with open(os.path.join(config.conversations_path, sanitizeName(subject) + config.fileExtention), "r") as f:
            chatRaw = f.read()
            bootstrappedChat = config.chat.bootStrapChat(config, scenario)
            chat = bootstrappedChat + "\n" + chatRaw  + enquiry + "\n" + config.progConfig["aiPrompt"]
            tries = 1
            if retry:
                tries = config.progConfig.get("maxRetries",1)
            success = False
            sleepBetweenRetries = config.progConfig["retryDelay"]
            while tries > 0:
                try:
                    response = config.chat.completions_with_backoff(
                        delay_in_seconds=config.delay,
                        model=model,
                        prompt=chat,
                        temperature=temperature,
                        max_tokens=max_tokens,
                        top_p=top_p,
                        frequency_penalty=frequency_penalty,
                        presence_penalty=presence_penalty,
                        stop=[ # "\n",
                        config.progConfig["userPrompt"], config.progConfig["aiPrompt"]],
                    )
                    ai = response.choices[0].text
                    if ai.startswith("\n\n"):
                        ai = ai[2:]
                    success = True
                    break
                except Exception as e:
                    tries -= 1

                    if str(e) == "openai.error.RateLimitError":
                        eprint("Error: Too many requests. We will try again")
                    eprint("Error: " + str(e))
                    eprint(f"Retrying again in {sleepBetweenRetries} seconds...")
                    time.sleep(sleepBetweenRetries)
                    sleepBetweenRetries *= config.progConfig["retryMultiplier"] 
                    if sleepBetweenRetries > config.progConfig["retryMaxDelay"]:
                        sleepBetweenRetries = config.progConfig["retryMaxDelay"]
        if success == False:
            eprint("Error: Too many requests. Please wait a few minutes and try again")
            return
        if save:               
            with open(os.path.join(config.conversations_path, sanitizeName(subject) + config.fileExtention), "a") as f:
                f.write(enquiry) 
                f.write("\n")
                f.write(config.progConfig["aiPrompt"] + ai)
                f.write("\n")
                f.close()
        if execute:
            editPrompt = click.prompt(f"{ai}\nedit command? [y/n]", type=click.Choice(["y", "n"]), default="n")
            if editPrompt == "y":
                edited = click.edit(ai)
                if edited:
                    ai = edited
            doExec = click.prompt(f"{ai}\nExecute command? [y/n]", type=click.Choice(["y", "n"]), default="y")
            """execute the command in the terminal and edit the response before saving it."""
            if doExec == "y":
                result  = subprocess.run(ai, stdout=subprocess.PIPE, shell=True, stderr=subprocess.STDOUT)
                result = result.stdout.decode("utf-8")
                print(result)
                saveOutput = click.prompt(f"save output? [Y/e/n]", type=click.Choice(["y", "e", "n"]), default="y")
                if saveOutput == "e":
                    edited = click.edit(result.stdout.decode("utf-8"))
                    if edited:
                        result = edited
                if saveOutput != "n":
                    with open(os.path.join(config.conversations_path, sanitizeName(subject) + config.fileExtention), "a") as f:
                        f.write(config.progConfig["userPrompt"] + str(result))
                        f.write("\n")
                        f.close()
        else:
            if verbose:
                print(chat + ai)
            else:
                print(ai)
    else:
        print("No subject provided")
        return
    
    #  print("The askGPT is sleeping. [bold red]Next time, rub the lamp first.[/bold red]")





import sys
import json

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)


def sanitizeName(name):
    """
    Sanitize the name of the conversation to be saved."""
    return name.replace(" ", "_").replace("/", "_")

def load_json(file):
    """
    Load json from file"""
    with open(file, "r") as f:
        try:
            return json.load(f)
        except:
            return dict()